/*!---------------------------------------------------------------------------
** @file	main.c
**
** @brief	Firmware main file for USB Modbus I/O Interface
**
**---------------------------------------------------------------------------
**
** Copyright 2010 Roy Hopkins (roy_web@fieldofcows.com)
**
** This file is part of the firmware for the Field of Cows USB Modbus I/O
** project (USBModbus)
** 
** USBModbus is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, either version 3 of the License, or
** (at your option) any later version.
** 
** USBModbus is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with USBModbus.  If not, see <http://www.gnu.org/licenses/>.
**
** Modified by Istvn Tuba (egyujabbgepepito)
** Added: 1.	2 MPG (Rol-over) counter
**	      2.	4 analog input
**		  3.	MBFN_WRITE_MULTIPLE_REGISTERS
**		  4.	2 PWM output
**		  5.	Tested with MACH3
**		  6.	LCD display support
*/
/*---------------------------------------------------------------------------
** Includes
*/

#include <p18f67j50.h>
#include <portb.h>

#include "usb_config.h"
#include "USB/usb.h"
#include "USB/usb_function_cdc.h"
#include "USB/usb_device.h"



#include "modbus.h"



/*---------------------------------------------------------------------------
** Defines and Macros
*/
#pragma config PLLDIV   = 1         	// (4 MHz crystal)
#pragma config CPUDIV   = OSC2_PLL2
#pragma config FOSC     = HSPLL
#pragma config FCMEN    = OFF
#pragma config IESO     = OFF
#pragma config WDTEN    = OFF
#pragma config STVREN   = ON
#pragma config XINST    = OFF       	// Extended Instruction Set
#pragma config CP0      = OFF
#pragma config MSSPMSK  = MSK5
#pragma config CCP2MX   = DEFAULT


#define CLOCK_FREQ 48000000
#define GetSystemClock() (CLOCK_FREQ)

	unsigned short led_count = 0;
/*---------------------------------------------------------------------------
** Typedefs
*/
/*---------------------------------------------------------------------------
** Local function prototypes
*/
void handle_high_isr(void);
void handle_low_isr(void);
void BlinkUSBStatus(void);

/*---------------------------------------------------------------------------
** Data
*/

/*---------------------------------------------------------------------------
** Interrupt vectors and handlers
*/
  
//
// Define the high priority interrupt vector jump at address 0x08
//
#pragma code HIGH_INTERRUPT_VECTOR = 0x1008
void
high_isr(
  void
) {
	// The code here is in a vector table so we have only got a few
	// bytes to play with. Therefore just jump to the isr
	_asm goto handle_high_isr _endasm
}

//
// Define the low priority interrupt vector jump at address 0x18
//
#pragma code LOW_INTERRUPT_VECTOR = 0x1018
void
low_isr(
  void
) {
	_asm goto handle_low_isr _endasm
}

#pragma code 

//
// The vector jump for the high priority interrupt jumps
// to this function. The #pragma interrupt ensures that
// the correct return is generated from the function
//
#pragma interrupt handle_high_isr
void
handle_high_isr(
  void
) {
	// Handle any pending USB tasks
	USBDeviceTasks();
}

//
// The vector jump for the low priority interrupt jumps
// to this function. The #pragma interruptlow ensures that
// the correct return is generated from the function
//
#pragma interruptlow handle_low_isr
void
handle_low_isr(
  void
) {

	MPGProcess();

}

//--------------------------------------------------------------------------
/*!
** @brief   Main firmware process entry point
**
*/
void main(void) {
	unsigned char count = 0;


#define mInitAllLEDs()      LATF &= 0xFB; TRISF &= 0xFB;
#define mLED_1              LATFbits.LATF2
#define mLED_1_On()         mLED_1 = 1;
#define mLED_1_Off()        mLED_1 = 0;
#define mLED_1_Toggle()     mLED_1 = !mLED_1;

    {
        unsigned int pll_startup_counter = 600;
        OSCTUNEbits.PLLEN = 1;  //Enable the PLL and wait 2+ms until the PLL locks before enabling USB module
        while(pll_startup_counter--);
    }

	// Initialise the USB library
	mInitAllLEDs();
	mLED_1_On();

	USBDeviceInit();
	USBDeviceAttach();

	// Initialise the Modbus routines
#ifdef MODECONTROL_SW
	setMode(MODE_RTU);
#endif
	initialiseModbus();

	// Enter the main processing loop
	while (1) {
		if ((USBDeviceState >= CONFIGURED_STATE) &&
		    (USBSuspendControl != 1)) {
			processModbus();
			CDCTxService();
			BlinkUSBStatus();
		}
	}
}

//--------------------------------------------------------------------------
/*!
** @brief   Handler that is called when the host sets the baud rate for
**			the CDC connection
**
** This function gets called when a SetLineCoding command
** is sent on the bus.  This function will evaluate the request
** and determine if the application should update the baudrate
** or not.
**
*/
void
set_line_coding_handler(
	void
) {
	//If the request is not in a valid range
	if (cdc_notice.GetLineCoding.dwDTERate.Val > 115200) {
		// Stalling the endpoint is the correct thing to do in this
		// case but it can cause the host application to crash if it
		// doesn't handle this properly. If that is the case then
		// the line below can be commented-out
		USBStallEndpoint(0, 1);
	}
	else {
		DWORD_VAL baud;

		// Update the baudrate info in the CDC driver
		CDCSetBaudRate(cdc_notice.GetLineCoding.dwDTERate.Val);

		// Update the baudrate of the UART
		baud.Val = (GetSystemClock() / 4) / line_coding.dwDTERate.Val - 1;
		SPBRG = baud.v[0];
		SPBRGH = baud.v[1];
	}
}

//--------------------------------------------------------------------------
/*!
** @brief		USB event callback
**
** @return 	TRUE
**
** This function is called from the USB stack to
** notify a user application that a USB event
** occured.  This callback is in interrupt context
**
** @param[in]	event	Type of event
** @param[in]	pdata	Pointer to event data
** @param[in]	size	Size of event data
*/
BOOL
USER_USB_CALLBACK_EVENT_HANDLER(
	USB_EVENT event,
 	void *pdata,
  	WORD size
) {
	// We just forward these events to the relevant CDC handler
	switch (event) {
		case EVENT_CONFIGURED:
			CDCInitEP();
			break;
		case EVENT_EP0_REQUEST:
			USBCheckCDCRequest();
			break;
		case EVENT_TRANSFER:
			Nop();
			break;
		default:
			break;
	}
	return TRUE;
}

//-----------------------------------------------------------------------------
void BlinkUSBStatus(void)
{
    if(led_count == 0)led_count = 10000U;
    led_count--;

            if(led_count==0)
            {
                mLED_1_Toggle();
            }

}